/*
 * Convert a ROM file to TI-83+/84+ 8xu
 *
 * Copyright (C) 2010 Benjamin Moody
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <rabbitsign.h>

static void *xmalloc(unsigned long n)
{
  void *p = malloc(n);

  if (!p) {
    fprintf(stderr, "out of memory (need %lu bytes)\n", n);
    abort();
  }

  return p;
}

static int read_page(RSProgram *prgm, int pagenum, FILE *infile, int hwpages)
{
  int realpagenum;
  unsigned char buf[16384];
  int n;

  if (pagenum < 0x10)
    realpagenum = pagenum;
  else
    realpagenum = pagenum + hwpages - 0x20;

  fseek(infile, realpagenum * 0x4000L, SEEK_SET);
  n = fread(buf, 1, 16384, infile);
  if (n != 16384) {
    fprintf(stderr, "I/O error\n");
    return 1;
  }

  if (pagenum == 0) {
    buf[0x0056] = 0xff;
  }

  if (pagenum == 0x1A) {
    if (buf[256] >= 0xfe) {
      fprintf(stderr, "warning: no OS signature found\n");
    }
    else {
      memset(prgm->signature, 0xff, 256);
      prgm->signature[0] = 0x02;
      prgm->signature[1] = 0x0d;
      memcpy(prgm->signature + 2, buf + 256, buf[256] + 1);
      prgm->signature_length = buf[256] + 3;
      if (prgm->signature_length < 96)
	prgm->signature_length = 96;
    }

    memset(buf, 0xff, 512);
  }
  
  prgm->pagenums[prgm->npagenums] = pagenum;
  prgm->npagenums++;

  rs_program_append_data(prgm, buf, n);

  return 0;
}

static const char usage[] = "Usage: %s [-o output.8xu] input.rom\n";

int main(int argc, char **argv)
{
  const char *infilename = NULL;
  const char *outfilename = "out.8xu";
  long int length;
  RSProgram *prgm;
  FILE *infile;
  FILE *outfile;
  int hwpages;
  unsigned long hdrstart, hdrsize, fieldstart, npages;
  const unsigned char *hdr;
  int i;
  int status = 0;
  
  rs_set_progname(argv[0]);
  rs_set_verbose(0);

  for (i = 1; i < argc; i++) {
    if (argv[i][0] != '-')
      infilename = argv[i];
    else if (argv[i][1] == 'o') {
      if (argv[i][2])
	outfilename = &argv[i][2];
      else if (++i < argc)
	outfilename = argv[i];
      else {
	fprintf(stderr, usage, argv[0]);
	return 1;
      }
    }
    else {
      fprintf(stderr, "%s: unknown option %s\n", argv[0], argv[i]);
      fprintf(stderr, usage, argv[0]);
      return 1;
    }
  }

  if (!infilename) {
    fprintf(stderr, "%s: no input file specified\n", argv[0]);
    fprintf(stderr, usage, argv[0]);
    return 1;
  }

  infile = fopen(infilename, "rb");
  if (!infile) {
    perror(infilename);
    return 1;
  }

  fseek(infile, 0L, SEEK_END);
  length = ftell(infile);

  if (length >= 0x80000 && length <= 0x89000) {
    hwpages = 0x20;
  }
  else if (length >= 0x100000 && length <= 0x121000) {
    hwpages = 0x40;
  }
  else if (length >= 0x200000 && length <= 0x221000) {
    hwpages = 0x80;
  }
  else {
    fprintf(stderr, "%s: invalid ROM size\n", infilename);
    fclose(infile);
    return 1;
  }

  prgm = rs_program_new();
  prgm->calctype = RS_CALC_TI83P;
  prgm->datatype = RS_DATA_OS;

  prgm->header = xmalloc(256);
  prgm->signature = xmalloc(256);

  fseek(infile, (hwpages - 6) * 0x4000L, SEEK_SET);
  fread(prgm->header, 1, 256, infile);

  rs_get_field_size(prgm->header, &hdrstart, NULL);
  hdr = prgm->header + hdrstart;
  hdrsize = 256 - hdrstart;

  npages = rs_get_numeric_field(0x8080, hdr, hdrsize);

  if (!rs_find_app_field(0x8070, hdr, hdrsize, NULL, &fieldstart, NULL))
    prgm->header_length = hdrstart + fieldstart;
  else
    prgm->header_length = 256;

  if (npages == 0)
    npages = 14;

  if (npages > 22)
    npages = 22;

  prgm->pagenums = xmalloc(npages * sizeof(unsigned int));

  status = read_page(prgm, 0, infile, hwpages);
  if (!status && npages >= 4)
    status = read_page(prgm, 1, infile, hwpages);
  if (!status && npages >= 5)
    status = read_page(prgm, 2, infile, hwpages);
  if (!status && npages >= 6)
    status = read_page(prgm, 3, infile, hwpages);

  if (!status && npages >= 11)
    status = read_page(prgm, 4, infile, hwpages);
  if (!status && npages >= 12)
    status = read_page(prgm, 5, infile, hwpages);
  if (!status && npages >= 13)
    status = read_page(prgm, 6, infile, hwpages);
  if (!status && npages >= 14)
    status = read_page(prgm, 7, infile, hwpages);

  if (!status && npages >= 19)
    status = read_page(prgm, 0x10, infile, hwpages);
  if (!status && npages >= 20)
    status = read_page(prgm, 0x11, infile, hwpages);
  if (!status && npages >= 21)
    status = read_page(prgm, 0x12, infile, hwpages);
  if (!status && npages >= 22)
    status = read_page(prgm, 0x13, infile, hwpages);

  if (!status && npages >= 15)
    status = read_page(prgm, 0x14, infile, hwpages);
  if (!status && npages >= 16)
    status = read_page(prgm, 0x15, infile, hwpages);
  if (!status && npages >= 17)
    status = read_page(prgm, 0x16, infile, hwpages);
  if (!status && npages >= 18)
    status = read_page(prgm, 0x17, infile, hwpages);

  if (!status && npages >= 7)
    status = read_page(prgm, 0x18, infile, hwpages);
  if (!status && npages >= 8)
    status = read_page(prgm, 0x19, infile, hwpages);
  if (!status && npages >= 9)
    status = read_page(prgm, 0x1A, infile, hwpages);
  if (!status && npages >= 10)
    status = read_page(prgm, 0x1B, infile, hwpages);

  if (!status && npages >= 2)
    status = read_page(prgm, 0x1C, infile, hwpages);
  if (!status && npages >= 3)
    status = read_page(prgm, 0x1D, infile, hwpages);

  fclose(infile);

  if (status) {
    rs_program_free(prgm);
    return 1;
  }

  outfile = fopen(outfilename, "wb");
  if (!outfile) {
    perror(outfilename);
    rs_program_free(prgm);
    return 1;
  }

  if (rs_write_ti8x_file(prgm, outfile, 0, 0, 0, 0)) {
    fclose(outfile);
    rs_program_free(prgm);
    return 1;
  }

  fclose(outfile);
  rs_program_free(prgm);
  return 0;
}
